/**
 * Copyright 2011-2012 Bull S.A.S.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.ow2.util.maven.plugin.mergeproperties.core;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.StringTokenizer;

import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;

/**
 * Merge properties files
 * @author Jeremy Cazaux
 */
public class MergeProperties {

    /**
     * The master resource
     */
    private Resource masterResource;

    /**
     * Template resources
     */
    private List<Resource> templateResources;

    /**
     * Properties resources
     */
    private List<Resource> propertiesResources;

    /**
     * Name of the output merge file
     */
    private String outputFile;

    /**
     * Strategies
     */
    private Strategies strategies;

    /**
     * Preferences
     */
    private Preferences preferences;

    /**
     * List of markers of a set of template files
     */
    private List<Marker> markers;

    /**
     * List of file order preferences
     */
    private List<ResourceOrderPreference> resourceOrderPreferences;

    /**
     * End of line
     */
    private static final String EOL = "\n";

    /**
     * Logger
     */
    private Log logger = LogFactory.getLog(this.getClass());

    /**
     * Default constructor
     */
    public MergeProperties() {
    }

    /**
     * @param key
     * @return the marker identified by his key if found, null otherwise
     */
    public  Marker getMarker(final List<Marker> markers, final String key) {
        int i = 0;
        //search of marker
        while (i < markers.size() && !markers.get(i).getKey().equals(key)) {
            i++;
        }

        //return the marker if found, null otherwise
        if (i < markers.size() && markers.get(i).getKey().equals(key)) {
            return markers.get(i);
        } else {
            return  null;
        }
    }

    /**
     * Test if a there is a marker which contain the key
     * @param key
     * @return true if a marker is identified by the key
     */
    public boolean isMarkersContainKey(final String key) {
        int i = 0;
        while (i < this.markers.size() && !this.markers.get(i).getKey().equals(key)) {
            i++;
        }
        return (i < this.markers.size());
    }

    /**
     * Sort resourceOrderPreferences list by priority and by the mostCommonSubKey
     */
    private void sortFileOrderPreferences(){
        Collections.sort(this.resourceOrderPreferences, new ResourceOrderPreferenceComparator());
    }

    /**
     * @param line A line of a property file
     * @return the StringTokenizer of the line
     */
    public StringTokenizer getStringTokenizer(String line)  {
        if (line.contains("=")) {
            return new StringTokenizer(line, "=");
        } else {
            return new StringTokenizer(line, " ");
        }
    }

    /**
     * Check if parameters are valid
     * @throws Exception
     */
    public void check() throws Exception {
        if (this.masterResource == null) {
            throw new Exception("Master resource is not set");
        }

        if (this.templateResources == null) {
            throw new Exception("Template resources are not set");
        }

        if (this.outputFile == null) {
            throw new Exception("Output file is not set");
        }
    }

    /**
     * Main method of the plugin
     * @throws Exception
     * @throws Exception
     */
    public void execute() throws Exception {

        BufferedReader br = null;

        this.check();

        List<Resource> templates = new ArrayList<Resource>();
        templates.add(this.masterResource);
        templates.addAll(this.templateResources);
        this.markers = new ArrayList<Marker>();

        //key and value of all properties
        //if there are several values for a same key, it's will be add to the marker list (markers variable)
        List<Marker> properties = new ArrayList<Marker>();

        //defaut strategy if no strategies are defined
        //the default strategy is an aggregate strategy
        ExclusiveStrategy defaultStrategy = new ExclusiveStrategy();

        //initialize SubKeyOccList with priority and pattern values
        List<SubsequenceKeyOccurrence> subsequenceKeyOccurences = new ArrayList<SubsequenceKeyOccurrence>();
        this.resourceOrderPreferences = new ArrayList<ResourceOrderPreference>();

        subsequenceKeyOccurences.add(new SubsequenceKeyOccurrence(0));
        if (this.preferences!=null){
            List<String> patterns = this.preferences.getFileOrderPreferences();
            int i=1;
            if (patterns!=null){
                for(String pattern: patterns){
                    subsequenceKeyOccurences.add(new SubsequenceKeyOccurrence(i,pattern));
                    i++;
                }
            }
        }

        try {
            //reading of template and properties files
            List<Resource> resources = new ArrayList<Resource>();
            resources.addAll(templates);
            if (this.propertiesResources != null) {
                resources.addAll(this.propertiesResources);
            }
            for (Resource resource: resources) {

                br = resource.getBufferedReader();

                String line;
                String key = null;
                String value = null;
                List<String> subKeys;

                //initialize subsequenceKeyOccurrence
                for (SubsequenceKeyOccurrence subsequenceKeyOccurrence: subsequenceKeyOccurences) {
                    subsequenceKeyOccurrence.initHashMap();
                }

                try {
                    while ((line = br.readLine()) != null) {

                        StringTokenizer stringTokenizer = getStringTokenizer(line);
                        if (stringTokenizer.hasMoreElements()){
                            key=stringTokenizer.nextElement().toString();
                            if (key.charAt(0) == '%'){
                                if (!key.endsWith("%")) {
                                    throw new Exception("Resource " + resource.getName() + "- Marker " +key + " is not valid");
                                }
                                key = key.substring(1, key.length()-1);  //get marker string between '%' characters
                                this.markers.add(new Marker(key));   //addition of a new marker with a key
                            } else if (stringTokenizer.hasMoreElements()) {
                                value = stringTokenizer.nextElement().toString();

                                if (key.charAt(0) != '#') {
                                    //add property to the properties list
                                    Marker property = getMarker(properties, key);
                                    if (property != null) {
                                        property.addValue(value);
                                    } else {
                                        property = new Marker(key);
                                        property.addValue(value);
                                        if (line.contains("=")) {
                                            property.setSeparator("=");
                                        } else {
                                            property.setSeparator("    ");
                                        }
                                        properties.add(property);
                                    }
                                }

                                Marker marker = getMarker(this.markers, key);
                                if (marker != null) {

                                    //update separator value
                                    String separator;
                                    if (marker.getSeparator() == null) {
                                        if (line.contains("=")) {
                                            marker.setSeparator("=");
                                        } else {
                                            marker.setSeparator("    ");
                                        }
                                    }
                                    marker.addValue(value);  //addition of a value to an existing marker
                                }
                            }

                            if (key.charAt(0)!='#'){
                                //Add all subsequences of the key to the SubsequenceKeyOccurrence object
                                for (SubsequenceKeyOccurrence subsequenceKeyOccurrence: subsequenceKeyOccurences) {
                                    subsequenceKeyOccurrence.addSubKeys(key);
                                }
                            }
                        }

                        if (line.contains("\\")){
                            while ((line = br.readLine()) != null && line.contains("\\"));
                        }
                    }
                } catch (Exception e) {
                    logger.error("Cannot read line.", e);
                }

                //search of a the first SubsequenceKeyOccurrence object which isn't empty
                //the first SubsequenceKeyOccurrence object has the highest priority
                int i = subsequenceKeyOccurences.size() - 1;
                while (i >= 0 && subsequenceKeyOccurences.get(i).getHashMapSubKeyOcc().isEmpty()) {
                    i--;
                }

                if (i >= 0) {

                    //addition of a new ResourceOrderPreference object with the priority of the file, its name and its mots
                    //common subsequence key
                    SubsequenceKeyOccurrence subsequenceKeyOccurrence = subsequenceKeyOccurences.get(i);
                    subsequenceKeyOccurrence.sortHashMap(SubsequenceKeyOccurrence.OrderType.DESC);
                    Iterator it = subsequenceKeyOccurrence.getHashMapSubKeyOcc().keySet().iterator();

                    int fileOrderPriority;
                    int subKeyOccPriority;
                    subKeyOccPriority = subsequenceKeyOccurrence.getPriority();
                    if (subKeyOccPriority == 0) {
                        fileOrderPriority = subKeyOccPriority;
                    } else {
                        fileOrderPriority = subsequenceKeyOccurences.size() - subKeyOccPriority;
                    }
                    String mostCommonSubKey = (String)it.next();

                    this.resourceOrderPreferences.add(new ResourceOrderPreference(fileOrderPriority,resource,
                            mostCommonSubKey));

                    //some traces
                    logger.debug("Resource: " + resource.getName() + " | priority: " + fileOrderPriority + " | mostCommonSubKey:" +
                            mostCommonSubKey + this.EOL);
                }
            }

            //sort resourceOrderPreferences list by priority and by the most common key subsequence
            sortFileOrderPreferences();

            //initialize patterns of preferences
            if (this.preferences != null) {
                this.preferences.initPatterns();
            }

            if (this.strategies != null) {
                //initialize patterns of strategies
                this.strategies.initPatterns();

                logger.debug("AggregateStrategies:");
                List<AggregateStrategy> aggregateStrategies = strategies.getAggregateStrategies();
                if (aggregateStrategies != null && !aggregateStrategies.isEmpty()) {
                    for (AggregateStrategy aggregateStrategy: aggregateStrategies) {
                        logger.debug(aggregateStrategy.getKey() + " " + aggregateStrategy.getPattern());
                    }
                } else {
                    if (aggregateStrategies == null) {
                        logger.debug("aggregateStrategies is null");
                    } else {
                        logger.debug("aggregateStrategies is empty");
                    }
                }
            } else {
                logger.debug("strategies is null");
            }

            //some traces
            logger.debug(this.EOL + this.EOL +this.EOL +"(FILES,PRIORITY,MOST_COMMON_SUB_SEQUENCE)");
            for (ResourceOrderPreference resourceOrderPreference : this.resourceOrderPreferences) {
                logger.debug("(" + resourceOrderPreference.getResource().getName() + "," + resourceOrderPreference.getPriority()
                        + "," + resourceOrderPreference.getMostCommonSubKey() + ")");
            }

            //update markers list with properties wich have several values for one property
            for (Marker property: properties) {
                if (property.getValues().size() > 1 && getMarker(this.markers, property.getKey()) == null) {

                    //remove dump values
                    List<String> values = property.getValues();
                    List<String> valuesToRemove = new ArrayList<String>();
                    for (int i = 0; i < values.size() - 1; i++) {
                        for (int y = i + 1; y < values.size(); y++) {
                            if (i != y && values.get(i).equals(values.get(y))) {
                                valuesToRemove.add(values.get(i));
                            }
                        }
                    }
                    for (String valueToRemove: valuesToRemove) {
                        values.remove(valueToRemove);
                    }

                    this.markers.add(property);
                }
            }

            //Log markers
            logger.debug("Markers: " + this.markers.size());
            for (Marker marker: this.markers) {
                String values = "";
                Iterator it = marker.getValues().iterator();

                if (it.hasNext()) {
                    while (it.hasNext()) {
                        values += "," + it.next();
                    }
                    logger.debug(marker.getKey() + marker.getSeparator() + values.subSequence(1, values.length()));
                }
            }

            BufferedWriter bw = null;
            Marker marker;
            boolean eol;
            try {

                //create parent directories if !exist
                File outputFile = new File(this.outputFile);
                File parent = outputFile.getParentFile();
                if (!parent.exists()) {
                    parent.mkdirs();
                }

                bw = new BufferedWriter(new FileWriter(outputFile));

                //writing of output file
                for (ResourceOrderPreference resourceOrderPreference : this.resourceOrderPreferences) {

                    String line = null;
                    String key = null;
                    eol = false;

                    logger.debug("Resource name == " + resourceOrderPreference.getResource().getName());

                    Resource resource = resourceOrderPreference.getResource();

                    try {
                        br = resource.getBufferedReader();

                        while ((line = br.readLine()) != null) {
                            StringTokenizer stringTockenizer = getStringTokenizer(line);
                            if (stringTockenizer.hasMoreElements()) {
                                key=stringTockenizer.nextElement().toString();

                                if (key.charAt(0) == '%') {
                                    //get marker string between '%' characters
                                    key = key.substring(1, key.length() - 1);
                                }

                                marker = getMarker(this.markers, key);

                                if (marker != null) {

                                    if (!marker.isWrite()) {

                                        //get values as a String list
                                        List<String> values;
                                        if (this.preferences == null) {
                                            values = marker.getValues();
                                        } else {
                                            //apply preferences
                                            values = this.preferences.applyPreferences(marker);
                                        }

                                        //apply strategies
                                        if (this.strategies != null) {
                                            line = this.strategies.execute(key, values, marker.getSeparator(),
                                                    defaultStrategy);
                                        } else {
                                            line = defaultStrategy.execute(key, values, marker.getSeparator());
                                        }

                                        //the marker has been written into the output file
                                        marker.setWrite(true);

                                    } else {
                                        line = null;
                                    }
                                }
                            }

                            if (line != null) {
                                bw.write(line + "\n");
                                if (eol == false)
                                    eol = true;
                            }
                        }
                    } catch (FileNotFoundException e) {
                        throw new Exception("Cannot find the resource '" +
                                resourceOrderPreference.getResource().getName() + "'.", e);
                    } catch (IOException e) {
                        throw new Exception("Cannot read line.",e);
                    }

                    try {
                        if (eol)
                            bw.write("\n");
                    } catch (IOException e) {
                        throw new Exception("Cannot write line into the buffered writer.", e);
                    }
                }

            } catch (IOException e) {
                throw new Exception("Cannot create the file '" + this.outputFile + "'.", e);
            } finally {
                try {
                    bw.close();
                } catch (IOException e) {
                    throw new Exception("Cannot close buffered writer.", e);
                }
            }

        } finally {
            //close all resources
            this.masterResource.close();
            for (Resource resource: templateResources) {
                resource.close();
            }

            if (propertiesResources != null) {
                for (Resource resource: propertiesResources) {
                    resource.close();
                }
            }
        }
    }

    /***
     *
     * @param masterResource Master resource to set
     */
    public void setMasterResource(Resource masterResource) {
        this.masterResource = masterResource;
    }

    /**
     *
     * @param templateResources Template resources to set
     */
    public void setTemplateResources(List<Resource> templateResources) {
        this.templateResources = templateResources;
    }

    /**
     *
     * @param propertiesResources Properties resources to set
     */
    public void setPropertiesResources(List<Resource> propertiesResources) {
        this.propertiesResources = propertiesResources;
    }

    /**
     *
     * @param outputFile Outputfile to set
     */
    public void setOutputFile(String outputFile) {
        this.outputFile = outputFile;
    }

    /**
     *
     * @param strategies Strategies to set
     */
    public void setStrategies(Strategies strategies) {
        this.strategies = strategies;
    }

    /**
     *
     * @param preferences Preferences to set
     */
    public void setPreferences(Preferences preferences) {
        this.preferences = preferences;
    }
}

