/**
 * Copyright 2011-2012 Bull S.A.S.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.ow2.util.maven.plugin.mergeproperties.core;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;

/**
 * This class represents preferences of the merge properties plugin
 * @author Jeremy Cazaux
 */
public class Preferences {

    /**
     * With a fixed preference, we associate a key only with values of the fixedPreference object
     * @parameter
     */
    private Preference[] fixedPreferences;

    /**
     * fixedPreferences as a List
     */
    private List<Preference> fixedPreferencesList;

    /**
     * With a order preference, values of the orderPreference object are priority.
     * @parameter
     */
    private Preference[] orderPreferences;

    /**
     * orderPreferences as a List
     */
    private List<Preference> orderPreferencesList;

    /**
     * fileOrderPreferences are pattern used to define priorities to:
     * - find the most common subsequence key of template or properties file.
     * - sort files in ascending order, firstly by pattern priority, then by most common subsequence key
     * Priorities are defined by ascending order
     *
     * @parameter
     */
    private List<String> fileOrderPreferences;

    /**
     * The logger
     */
    private static Log logger = LogFactory.getLog(Preferences.class);

    /**
     * Default constructor
     */
    public Preferences() {
    }

    /**
     * Apply preferences to a marker and return values as list
     * @param marker
     * @return values. If a fixed preference is found, return only values defined in fixed preference object.
     * If a order preference is found, return all values. Values are sorted
     */
    public List<String> applyPreferences(final Marker marker) throws Exception {
        List<String> validValues = new ArrayList<String>();
        List<String> preferencesValues;
        List<String> markerValues = marker.getValues();
        int i = 0;

        //looking for fixed preferences
        if (this.fixedPreferencesList != null){

            //search for fixed preferences
            while (i < this.fixedPreferencesList.size() && !this.fixedPreferencesList.get(i).matchPattern(marker.getKey())) {
                i++;
            }

            //apply fixed preference if found
            if (i < this.fixedPreferencesList.size() && this.fixedPreferencesList.get(i).matchPattern(marker.getKey())){
                preferencesValues = this.fixedPreferencesList.get(i).getListValues();
                for (String value: preferencesValues){
                    if (markerValues.contains(value)){
                        validValues.add(value);
                        markerValues.remove(value);
                    }
                }
            }
        }

        if (this.orderPreferencesList != null) {

            //search for order preference
            i = 0;
            while (i < this.orderPreferencesList.size() && !this.orderPreferencesList.get(i).matchPattern(marker.getKey())) {
                i++;
            }

            //apply order preferences if found  and if no other preference are defined with the same key
            if (i < this.orderPreferencesList.size() && this.orderPreferencesList.get(i).matchPattern(marker.getKey())) {

                if (validValues.isEmpty()) {
                    preferencesValues = this.orderPreferencesList.get(i).getListValues();
                    for (String value: preferencesValues) {
                        if (markerValues.contains(value)) {
                            validValues.add(value);
                            markerValues.remove(value);
                        }
                    }

                    //Addition of other values
                    Iterator it = markerValues.iterator();
                    while (it.hasNext()) {
                        validValues.add(it.next().toString());
                    }

                } else {
                    throw new Exception("There is more than one preference defined for key " + marker.getKey());
                }
            }
        }

        //If no preference are defined, addition of other values
        if (validValues.isEmpty()) {
            Iterator it = markerValues.iterator();
            while (it.hasNext()) {
                validValues.add(it.next().toString());
             }
        }

        return validValues;
    }

    /**
     *
     * @return fixed preferences as a list of Preference
     */
    public List<Preference> getFixedPreferences() {
        return this.fixedPreferencesList;
    }

    /**
     *
     * @return order preferences as a list of Preference
     */
    public List<Preference> getOrderPreferences() {
        return this.orderPreferencesList;
    }

    /**
     *
     * @return  file order preferences as a List of String
     */
    public List<String> getFileOrderPreferences() {
        return this.fileOrderPreferences;
    }

    /**
     * Initialize FixedPreferences and orderPreferences as a List and initialize pattern of fixed and order preferences
     */
    public void initPatterns() {

        if (this.fixedPreferences != null) {
            this.fixedPreferencesList = new java.util.ArrayList<Preference>(java.util.Arrays.asList(this.fixedPreferences));

                for (Preference fixPreference: this.fixedPreferencesList) {
                    fixPreference.initPattern();
                }
        }

        if (this.orderPreferences != null) {
            this.orderPreferencesList = new java.util.ArrayList<Preference>(java.util.Arrays.asList(this.orderPreferences));

                for (Preference ordPreference: this.orderPreferencesList) {
                    ordPreference.initPattern();
                }
        }
    }

    /**
     *
     * @param fixedPreferences Fixed preferences to set
     */
    public void setFixedPreferences(Preference[] fixedPreferences) {
        this.fixedPreferences = fixedPreferences;
    }

    /**
     *
     * @param orderPreferences Order preferences to set
     */
    public void setOrderPreferences(Preference[] orderPreferences) {
        this.orderPreferences = orderPreferences;
    }

    /**
     * @param fileOrderPreferences FileOrderPreferences to set
     */
    public void setFileOrderPreferences(List<String> fileOrderPreferences) {
        this.fileOrderPreferences = fileOrderPreferences;
    }
}
