/**
 * Copyright 2011-2012 Bull S.A.S.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.ow2.util.maven.plugin.mergeproperties.core;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

/**
 * This class represents all strategies to applied
 * @author Jeremy Cazaux
 */
public class Strategies {

    /**
     * ExclusiveStrategies as an Array
     * @parameter
     */
    private ExclusiveStrategy[] exclusiveStrategies;

    /**
     * ExclusiveStrategies as a List
     */
    private List<ExclusiveStrategy> exclusiveStrategiesList;

    /**
     * AggregateStrategies as an Array
     * @parameter
     */
    private AggregateStrategy[] aggregateStrategies;

    /**
     * AggregateStrategies as a List
     */
    private List<AggregateStrategy> aggregateStrategiesList;

    /**
     * Default constructor
     */
    public Strategies(){

    }

    /**
     * Execute all strategies associated to a key and values and return a property
     * @param key
     * @param values
     * @return
     * @throws Exception
     */
    public String execute(final String key, final List<String> values, final String separator,
                          final FilteringStrategy defaultStrategy) throws Exception {
        String property = null;
        int i = 0;

        //search for an aggregate strategy associate to the key and values
        if (this.aggregateStrategiesList != null){
            while (i < this.aggregateStrategiesList.size() && !this.aggregateStrategiesList.get(i).matchPattern(key)) {
                i++;
            }

            //execute an aggregate strategy if found
            if (i < this.aggregateStrategiesList.size() && this.aggregateStrategiesList.get(i).matchPattern(key)) {
                property = this.aggregateStrategiesList.get(i).execute(key, values, separator);
            }
        }

        //search for an exclusive strategy associate to the key and values
        if (property == null && this.exclusiveStrategiesList!=null) {
            i=0;
            while(i < this.exclusiveStrategiesList.size() && !this.exclusiveStrategiesList.get(i).matchPattern(key)) {
                i++;
            }

            //execute an exclusive strategy if found
            if (i < this.exclusiveStrategiesList.size() && this.exclusiveStrategiesList.get(i).matchPattern(key)) {
                property = this.exclusiveStrategiesList.get(i).execute(key,values, separator);
            }
        }

        //if no strategy is defined, execute the default strategy
        if (property == null) {
            return defaultStrategy.execute(key, values, separator);
        }

        return property;
    }

    /**
     *
     * @return the list of aggregate strategies
     */
    public List<AggregateStrategy> getAggregateStrategies() {
        return this.aggregateStrategiesList;
    }

    /**
     *
     * @return the list of exclusive strategies
     */
    public List<ExclusiveStrategy> getExclusiveStrategies() {
        return this.exclusiveStrategiesList;
    }

    /**
     *
     * @param exclusiveStrategies Exclusive strategies to set
     */
    public void setExclusiveStrategies(ExclusiveStrategy[] exclusiveStrategies) {
        this.exclusiveStrategies = exclusiveStrategies;
    }

    /**
     *
     * @param aggregateStrategies Aggregate strategies to set
     */
    public void setAggregateStrategies(AggregateStrategy[] aggregateStrategies) {
        this.aggregateStrategies = aggregateStrategies;
    }

    /**
     * Initialize AggregateStrategies and ExclusiveStrategies as a List and initialize pattern of preferences
     */
    public void initPatterns() {

        if (this.exclusiveStrategies != null) {

            this.exclusiveStrategiesList = new ArrayList(Arrays.asList(exclusiveStrategies));

            //initialize pattern of exclusive strategies
            for (ExclusiveStrategy exclusiveStrategy: this.exclusiveStrategiesList) {
                exclusiveStrategy.initPattern();
            }
        }

        if (this.aggregateStrategies != null) {

            this.aggregateStrategiesList = new ArrayList<AggregateStrategy>(Arrays.asList(aggregateStrategies));

            //initialize pattern of aggregate strategies
            for (AggregateStrategy aggregateStrategy: this.aggregateStrategiesList) {
                aggregateStrategy.initPattern();
            }
        }
    }

}
