/**
 * Copyright 2011-2012 Bull S.A.S.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.ow2.util.maven.plugin.mergeproperties.core;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.regex.Pattern;

/**
 * This class represents all subsequence of keys and their occurrence associated to a file and to a priority
 * @author Jeremy Cazaux
 */
public class SubsequenceKeyOccurrence {

    /**
     * Highter the priority is, higher the object is classified
     */
    private int priority;

    /**
     * Pattern associated to fileOrderPreference parameter
     */
    private Pattern pattern;

    /**
     * HashMap of (key,occurrence)
     */
    private HashMap<String,Integer> hashMapSubKeyOcc;

    /**
     * Order type
     */
    public static enum OrderType{ASC,DESC};


    /**
     * Constructor
     * @param priority
     * @param pattern
     */
    public SubsequenceKeyOccurrence(final int priority, final String pattern) {
        this.priority = priority;
        this.pattern = Pattern.compile(pattern);
        this.hashMapSubKeyOcc = new HashMap<String,Integer>();
    }

    /**
     * Constructor
     * @param priority
     */
    public SubsequenceKeyOccurrence(final int priority) {
        this.priority = priority;
        this.hashMapSubKeyOcc = new HashMap<String,Integer>();
        this.pattern = null;
    }

    /**
     * @return the priority
     */
    public int getPriority() {
        return this.priority;
    }

    /**
     *
     * @param priority The priority is set
     */
    public void setPriority(final int priority) {
        this.priority = priority;
    }

    /**
     * Initialize the hashMap of (key, occurrence)
     */
    public void initHashMap() {
        this.hashMapSubKeyOcc = new HashMap<String, Integer>();
    }

    /**
     * Test if the string match the pattern
     * @param key
     * @return true if s match pattern; false otherwise.
     */
    private boolean matchPattern(final String key) {
        return this.pattern.matcher(key).matches();
    }

    /**
     * @param s1
     * @param s2
     * @return the longest common subsequence of s1 and s2
     */
    public String getLongestCommonSubSequence(final String s1, final String s2) {
        int i = 0;
        while (i < s1.length() && i < s2.length() && s1.charAt(i) == s2.charAt(i)) {
            i++;
        }
        return s1.substring(0, i);
    }

    /**
     * Add all subsequences of the key to the HashMap hashMapSubKeyOcc
     * @param key
     */
    public void addSubKeys(final String key) {
        List<String> subKeys = new ArrayList<String>();
        Set<String> keys = this.hashMapSubKeyOcc.keySet();
        Iterator it = keys.iterator();
        String k;
        String commonSubSequence;

        //test if the pattern is defined and if the key match the pattern
        if (this.pattern == null || (this.pattern != null && matchPattern(key))) {

            //search of common subsequences of keys between the parameter and all key of the hashmap
            while (it.hasNext()) {
                k = (String) it.next();
                commonSubSequence = getLongestCommonSubSequence(k,key);
                if (!commonSubSequence.equals("") && !subKeys.contains(commonSubSequence)) {
                    subKeys.add(commonSubSequence);
                }
            }

            /**
             * add common subsequences of keys between the parameter and all key of the hashmap  to the hashmap
             */
            for (String subKey:subKeys) {
                if (this.hashMapSubKeyOcc.containsKey(subKey)) {
                    this.hashMapSubKeyOcc.put(subKey, this.hashMapSubKeyOcc.get(subKey) + 1);
                } else {
                    this.hashMapSubKeyOcc.put(subKey, 2);
                }
            }

            /**
             * Add the key to the hashmap
             */
            if (this.hashMapSubKeyOcc.containsKey(key)) {
                this.hashMapSubKeyOcc.put(key, this.hashMapSubKeyOcc.get(key) + 1);
            } else {
                this.hashMapSubKeyOcc.put(key, 1);
            }
        }
    }

    /**
     *  Sort the hashmap hashMapSubKeyOcc by orderType
     * @param orderType
     */
    public void sortHashMap(final OrderType orderType)  {

        List list = new LinkedList(this.hashMapSubKeyOcc.entrySet());
        Collections.sort(list, new Comparator() {
            public int compare(Object o1, Object o2) {
                if (orderType.equals(OrderType.DESC)) {
                    return ((Comparable) ((Map.Entry) (o2)).getValue()).compareTo(((Map.Entry) (o1)).getValue());
                } else {
                    return ((Comparable) ((Map.Entry) (o1)).getValue()).compareTo(((Map.Entry) (o2)).getValue());
                }
            }
        });

        Map result = new LinkedHashMap();
        for (Iterator it = list.iterator(); it.hasNext();) {
             Map.Entry entry = (Map.Entry)it.next();
             result.put(entry.getKey(), entry.getValue());
        }

        this.hashMapSubKeyOcc = (HashMap<String, Integer>) result;
    }

    /**
     * @return the HashMap of (key,occurrence)
     */
    public HashMap<String, Integer> getHashMapSubKeyOcc() {
        return this.hashMapSubKeyOcc;
    }
}
