/**
 * Copyright 2011-2012 Bull S.A.S.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.ow2.util.maven.plugin.mergeproperties.mojo;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import org.apache.maven.plugin.AbstractMojo;
import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.plugin.MojoFailureException;
import org.apache.maven.plugins.annotations.Mojo;
import org.apache.maven.plugins.annotations.Parameter;
import org.ow2.util.maven.plugin.mergeproperties.core.MergeProperties;
import org.ow2.util.maven.plugin.mergeproperties.core.Preferences;
import org.ow2.util.maven.plugin.mergeproperties.core.Resource;
import org.ow2.util.maven.plugin.mergeproperties.core.Strategies;

/**
 * @author Jeremy Cazaux
 * @since 2.0.0
 */
@Mojo(name = "merge")
public class MergePropertiesFilesMojo extends AbstractMojo {

    /**
     * Name of the master template file
     * @since 2.0.0
     */
    @Parameter(required = true)
    private String masterTemplateFile;

    /**
     * Name of the directory that contain all template files
     * @since 2.0.0
     */
    @Parameter(required = true)
    private String templatesDirectory;

    /**
     * Name of the directory that contain all properties files
     * @since 2.0.0
     */
    @Parameter
    private String propertiesDirectory;

    /**
     * Name of the output merge file
     * @since 2.0.0
     */
    @Parameter(required = true)
    private String outputFile;

    /**
     * Strategies
     * @since 2.0.0
     */
    @Parameter(required = true)
    private Strategies strategies;

    /**
     * Preferences
     * @since 2.0.0
     */
    @Parameter(required = true)
    private Preferences preferences;

    /**
     * Main method of the plugin
     * @throws MojoExecutionException
     * @throws MojoFailureException
     */
    public void execute() throws MojoExecutionException, MojoFailureException {

        FileInputStream fileInputStream;
        BufferedInputStream bufferedInputStream;

        //load master file
        try {
            fileInputStream = new FileInputStream(this.masterTemplateFile);
        } catch (FileNotFoundException e) {
            throw new MojoExecutionException("Cannot find the file '" + this.masterTemplateFile + "'.", e);
        }

        Resource masterResource = new Resource(fileInputStream);
        masterResource.setName(this.masterTemplateFile);

        //load template files ;
        List<Resource> templateResources = new ArrayList<Resource>();
        templateResources.addAll(getResourcesFromDirectory(this.templatesDirectory));

        //load properties files
        List<Resource> propertiesResources = new ArrayList<Resource>();
        propertiesResources.addAll(getResourcesFromDirectory(this.propertiesDirectory));

        //execute the merge properties
        MergeProperties mergeProperties = new MergeProperties();
        mergeProperties.setMasterResource(masterResource);
        mergeProperties.setTemplateResources(templateResources);
        mergeProperties.setPropertiesResources(propertiesResources);
        mergeProperties.setStrategies(this.strategies);
        mergeProperties.setPreferences(this.preferences);
        mergeProperties.setOutputFile(this.outputFile);
        try {
            mergeProperties.execute();
        } catch (Exception e) {
            throw new MojoExecutionException("",e);
        }
    }

    /**
     *
     * @param directory The directory which contain resources
     * @return Resources
     */
    List<Resource> getResourcesFromDirectory(String directory) throws MojoExecutionException {
        List<Resource> resources = new ArrayList<Resource>();

        if (directory !=  null && !directory.equals("")) {
            File dir = new File(directory);
            File[] filesAsArray = dir.listFiles();
            if (filesAsArray != null) {
                List<File> files = Arrays.asList(filesAsArray);
                for (File file: files) {
                    BufferedInputStream bufferedInputStream = null;
                    FileInputStream fileInputStream = null;
                    try {
                        fileInputStream = new FileInputStream(file);
                    } catch (FileNotFoundException e) {
                        throw new MojoExecutionException("Cannot find the file '" + file.getAbsolutePath() + "'.", e);
                    }

                    Resource resource = new Resource(fileInputStream);
                    resource.setName(file.getAbsolutePath());
                    resources.add(resource);

                    try {
                        fileInputStream.close();
                    } catch (IOException e) {
                        getLog().error("Cannot close FileInputStream of the resource" + file.getAbsolutePath());
                    }
                }
            }
        }
        return resources;
    }

}

